import argparse
from datetime import datetime

import numpy
from loguru import logger
import json
import time
import os
import copy
import cv2
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import csv

import sklearn.metrics
from sklearn.model_selection import ParameterGrid
from tqdm import tqdm
from shapely.geometry.linestring import LineString
from skimage.transform import hough_line, hough_line_peaks

from modules.file_utils import FileUtils
from modules.dict_to_obj import DictToObj
from modules.ply_reader import PlyReader
from modules.hough_transform import HoughTransform
from modules.helper_scripts import lines_to_features, features_to_lines, hex_to_rgb, generate_colours, fill_polygon
from modules.lines_utils import *
from modules.map_utils import *


def plot_poly(img, poly, path_vectors=np.array([]), show=False):
    img = copy.copy(img)
    for idx, point in enumerate(poly):
        p1 = [int(i) for i in point.as_list]
        if idx + 1 == len(poly):
            p2 = [int(i) for i in poly[0].as_list]
        else:
            p2 = [int(i) for i in poly[idx+1].as_list]
        cv2.line(img, tuple(p1), tuple(p2), (255, 0, 0), 3)

    if path_vectors.size != 0:
        for idx in range(len(path_vectors)-1):
            cv2.arrowedLine(img, (int(path_vectors[idx, 0]), int(path_vectors[idx, 1])),
                            (int(path_vectors[idx+1, 0]), int(path_vectors[idx+1, 1])),
                            thickness=1, color=(0, 255, 0))

    if show:
        cv2.imshow("Poly image", img)
        cv2.waitKey(0)  # cv2.imshow won't work without this :)
        cv2.destroyAllWindows()
    return img


def plot_from_midpoint(args, img, centers):
    fig, ax = plt.subplots(1,1)
    canvas = fig.canvas
    ax.axis('off')

    ax.imshow(img)
    length = 500

    if args.is_probabalistic_hough:
        ax.scatter(centers[:,0], centers[:,1], s=5, c='r')
        for center in centers:
            starty = center[1] - length * np.sin(center[2])
            startx = center[0] - length * np.cos(center[2])
            endy = center[1] + length * np.sin(center[2])
            endx = center[0] + length * np.cos(center[2])
            ax.plot([startx, endx], [starty, endy], c='g')
    else:
        for center in centers:
            a = np.cos(center[1])
            b = np.sin(center[1])
            x0 = a * center[0]
            y0 = b * center[0]
            x1 = int(x0 + length*(-b))
            y1 = int(y0 + length*(a))
            x2 = int(x0 - length*(-b))
            y2 = int(y0 - length*(a))
            ax.plot([x1, x2], [y1, y2], c='g')
    plt.show(block=False)
    a = 0


def line_approximation(args, data):
    if args.line_approx_method == 'mean':
        return np.mean(data, axis=0)
    if args.line_approx_method == 'median':
        return np.median(data, axis=0)
    if args.line_approx_method == 'medoid':
        pairwise_distances = sklearn.metrics.pairwise_distances(X=data)
        medoid_idx = np.argmin(pairwise_distances.sum(axis=0))
        return data[medoid_idx]


def lines_to_mask(args, map_image, lines):
    if map_image.ndim == 2:
        mask = np.zeros((map_image.shape), dtype=float)
    else:
        mask = np.zeros((map_image.shape[:2]), dtype=float)

    for line in lines:
        angle = line.angle
        p0 = (line.start_point.x + args.metric_line_width * np.cos(angle + np.pi/2), line.start_point.y + args.metric_line_width * np.sin(angle + np.pi/2))
        p1 = (line.start_point.x - args.metric_line_width * np.cos(angle + np.pi/2), line.start_point.y - args.metric_line_width * np.sin(angle + np.pi/2))
        p2 = (line.end_point.x - args.metric_line_width * np.cos(angle + np.pi/2), line.end_point.y - args.metric_line_width * np.sin(angle + np.pi/2))
        p3 = (line.end_point.x + args.metric_line_width * np.cos(angle + np.pi/2), line.end_point.y + args.metric_line_width * np.sin(angle + np.pi/2))
        coords = [p0, p1, p2, p3, p0]
        fill_polygon(coords, mask)
    return mask


def loss(gt_img, pred_img):
    return np.mean(np.abs(pred_img-gt_img))


def process_data(args, ply_filepath, path_filepath, use_45deg_lines):
    # filename = os.path.basename(ply_filepath).strip('.ply')  # TODO get filename from filepath

    map_data = Map(args, ply_filepath=ply_filepath, path_filepath=path_filepath, use_45deg_lines=use_45deg_lines)

    # preprocess map
    map_data.map_image = HoughTransform.detect_blobs(map_data.get_map(), args, show=False)

    # detect lines
    lines = HoughTransform.detect_lines(map_data.map_image, args)  # detect hough lines

    # probabilistic lines processing
    logger.info(f'Lines found {len(lines)}')
    lines_lengths = [l.transform(map_data.pix_to_meters).length for l in lines]
    # order descending
    lines = list(np.array(lines)[np.argsort(lines_lengths)[::-1]])
    lines_lengths = [l.transform(map_data.pix_to_meters).length for l in lines]
    lines = list(np.array(lines)[np.where(np.array(lines_lengths) > args.min_line_length)])
    logger.info(f'Lines found {len(lines)}')
    # find walls from lines
    map_data.process_and_add_lines(lines)

    # # save image to file
    # map_data.save_image(os.path.join(args.path_data_out, filename + '_lines_detected.jpg'), draw_lines=True, lines=lines)
    # map_data.save_image(os.path.join(args.path_data_out, filename + '_lines_processed.jpg'), draw_lines=True, draw_path=True)
    # map_data.save_map(os.path.join(args.path_data_out, filename + '_map.jpg'))

    # find collisions with path
    if path_filepath is not None:
        map_data.process_path()

    # map_data.save_image(os.path.join(args.path_data_out, filename + '_lines_path_processed.jpg'), draw_lines=True, draw_path=True)
    # map_data.save_map(os.path.join(args.path_data_out, filename + '_map_after_path.jpg'))
    # map_data.save_to_h5(os.path.join(args.path_data_out, filename + '.h5'))

    return map_data


def gridsearch():
    # gridsearch over parameters
    assert isinstance(args.path_data_in, str), logger.exception('args.path_data_in should be string')
    assert isinstance(args.path_data_out, str), logger.exception('args.path_data_out should be string')

    args_with_multiple_values = {}
    for key, value in args.__dict__.items():
        if isinstance(value, list):
            if len(value) > 1:
                args_with_multiple_values[key] = value

    for key in args.__dict__:
        if isinstance(args.__dict__[key], list):
            args.__dict__[key] = args.__dict__[key][0]

    csv_file = os.path.join(args.path_data_out, 'results.csv')
    grid_runs = list(ParameterGrid(args_with_multiple_values))
    logger.info(f'starting gridsearch with {len(grid_runs)} combinations')
    is_csv_header = False
    for idx, grid_each in enumerate(tqdm(grid_runs)):
        try:
            logger.info(f'Processing run {idx} with args - {json.dumps(grid_each, indent=4)}')
            args_local = copy.deepcopy(args.__dict__)
            args_local.update(grid_each)
            args_local = DictToObj(**args_local)
            args_local.path_data_out = os.path.join(args_local.path_data_out, f'{idx}')
            FileUtils.createDir(args_local.path_data_out)
            results = process_data(args_local)

            # state = {'idx': idx}
            # state.update(grid_each)
            # for key in results:
            #     for key2 in results[key]:
            #         state[key+' '+key2] = results[key][key2]
            # if not is_csv_header:
            #     with open(csv_file, 'w', newline='') as fp:
            #         writer = csv.DictWriter(fp, fieldnames=list(state.keys()))
            #         writer.writeheader()
            #         writer.writerow(state)
            #     is_csv_header = True
            # else:
            #     with open(csv_file, 'a', newline='') as fp:
            #         writer = csv.DictWriter(fp, fieldnames=list(state.keys()))
            #         writer.writerow(state)
        except Exception as e:
            logger.exception(e)


def main(ply_filepath, path_filepath, use_45deg_lines):
    arg_parser = argparse.ArgumentParser()
    # arg_parser.add_argument('-path_data_in', default=r"C:\Users\vecin\Documents\PycharmProjects\asya\alarm_pointcloud_backend_2021_Q4\data\input\ply", type=str)
    # arg_parser.add_argument('-path_data_out', default=r"C:\Users\vecin\Documents\PycharmProjects\asya\alarm_pointcloud_backend_2021_Q4\data_out", type=str)
    # arg_parser.add_argument('-is_delete_previous_output', default=False, type=lambda x: (str(x).lower() == 'true'))
    # arg_parser.add_argument('-path_data_gt', default=r'C:\Users\vecin\Documents\PycharmProjects\asya\alarm_pointcloud_backend_2021_Q4\data\gt', type=str)
    arg_parser.add_argument('-use_matplotlib_agg', default=False, type=lambda x: (str(x).lower() == 'true'))

    # map args
    arg_parser.add_argument('-pix_per_meter', default=50, type=int)
    arg_parser.add_argument('-map_padding', default=0.2, type=float) # part of map sizye
    arg_parser.add_argument('-wall_width', default=0.2, type=float) # in meters
    arg_parser.add_argument('-drone_width', default=0.3, type=float) # in meters
    arg_parser.add_argument('-use_45deg_lines', default=False, type=lambda x: (str(x).lower() == 'true')) # if true will search diagonal lines else only 90 deg lines

    # line processing args
    arg_parser.add_argument('-max_wall_gap', default=0.4, type=float) # in meters
    arg_parser.add_argument('-line_extend_step', default=0.1, type=float) # in meters
    arg_parser.add_argument('-min_line_distance', default=0.3, type=float) # in meters
    arg_parser.add_argument('-min_line_length', default=0.4, type=float) # in meters
    arg_parser.add_argument('-max_shift_after_rotate', default=0.2, type=float) # in meters
    arg_parser.add_argument('-min_pixel_density', default=0.2, type=float) # range 0 - 1

    # Hough transform params
    arg_parser.add_argument('-is_probabilistic_hough', default=True, type=lambda x: (str(x).lower() == 'true'))
    arg_parser.add_argument('-hough_rho', default=1, nargs='*', type=float)
    arg_parser.add_argument('-hough_theta', default=np.pi / 180, nargs='*', type=float)
    # higher threshold is needed for non-probabilistic Hough transform (40 for the first map)
    arg_parser.add_argument('-hough_threshold', default=10, nargs='*', type=int)
    arg_parser.add_argument('-hough_minLineLength', default=5, nargs='*', type=int)
    arg_parser.add_argument('-hough_maxLineGap', default=10, nargs='*', type=int)

    # Canny detector params
    arg_parser.add_argument('-canny_threshold1', default=50, nargs='*', type=int)
    arg_parser.add_argument('-canny_threshold2', default=200, nargs='*', type=int)
    arg_parser.add_argument('-canny_aperatureSize', default=3, nargs='*', type=int)  # usually between 3 and 7 (kernel size)
    arg_parser.add_argument('-canny_L2grad', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Blob detection params (default params detect single points)
    arg_parser.add_argument('-blobs_thresholdStep', default=1, nargs='*', type=int)  # don't change this
    arg_parser.add_argument('-blobs_minThreshold', default=127, nargs='*', type=int)  # don't change this
    arg_parser.add_argument('-blobs_maxThreshold', default=129, nargs='*', type=int)  # don't change this
    arg_parser.add_argument('-blobs_minDistBetweenBlobs', default=10.0, nargs='*', type=float)
    arg_parser.add_argument('-blobs_minArea', default=1, nargs='*', type=float)
    arg_parser.add_argument('-blobs_maxArea', default=10, nargs='*', type=float)
    arg_parser.add_argument('-blobs_minCircularity', default=0.7, nargs='*', type=float)
    arg_parser.add_argument('-blobs_minConvexity', default=0.7, nargs='*', type=float)
    arg_parser.add_argument('-blobs_minInertiaRatio', default=0.7, nargs='*', type=float)

    args, args_other = arg_parser.parse_known_args()
    logger.info('Map processing script starting')
    logger.info(json.dumps(args.__dict__))

    # if args.is_delete_previous_output:
    #     FileUtils.deleteDir(args.path_data_out)
    # FileUtils.createDir(args.path_data_out)

    if args.use_matplotlib_agg:
        matplotlib.use('Agg')

    return process_data(args, ply_filepath, path_filepath, use_45deg_lines)
