import os
import pathlib
import json
import threading
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog
from typing import Dict

from loguru import logger
import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseButton
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.image as mpimg
from matplotlib.offsetbox import TextArea, DrawingArea, OffsetImage, AnnotationBbox
import pygubu
from pygubu.builder.widgets.dialog import Dialog

from modules.dict_to_obj import DictToObj
from modules.map_utils import Map
from backend import main
from modules.lines_utils import *
from multiprocessing.pool import ThreadPool
from tkinter import DISABLED, LEFT

from modules.plt_draggable_point import PltDraggablePoint

TOOL_VIEWER = 'TOOL_VIEWER'
TOOL_LINES = 'TOOL_LINES'
TOOL_POI = 'TOOL_POI'

SUB_TOOL_POI_TARGET = 'SUB_TOOL_POI_TARGET'
SUB_TOOL_POI_STATION = 'SUB_TOOL_POI_STATION'
SUB_TOOL_POI_DRONE = 'SUB_TOOL_POI_DRONE'

SUB_TOOL_LINES_ADD = 'SUB_TOOL_LINES_ADD'
SUB_TOOL_LINES_MODIFY = 'SUB_TOOL_LINES_MODIFY'

UI_LABEL_MAP_ADD = 'Add first point with a click\n'\
                    'then add second point\n'\
                    'then move points by click and drag'

UI_LABEL_MAP_MODIFY = 'Select line you want to modify\n'\
                    'then move points by click and drag'

class DialogLoader:

    def __init__(self):
        self.builder = pygubu.Builder()
        self.builder.add_from_file('dialog_loader.ui')
        self.uiDialog: pygubu.builder.widgets.dialog.Dialog = self.builder.get_object('uiDialog')

        self.uiButtonOpenOk: ttk.Button = self.builder.get_object('uiButtonOk')
        self.uiButtonCancel: ttk.Button = self.builder.get_object('uiButtonCancel')
        self.uiButtonOpenPLY: ttk.Button = self.builder.get_object('uiButtonOpenPLY')
        self.uiButtonOpenTXT: ttk.Button = self.builder.get_object('uiButtonOpenTXT')

        self.uiCheckboxTXT: ttk.Checkbutton = self.builder.get_object('uiCheckboxTXT')
        self.uiCheckboxPLY: ttk.Checkbutton = self.builder.get_object('uiCheckboxPLY')

        self.uiCheckbox45Deg: ttk.Checkbutton = self.builder.get_object('uiCheckbox45Deg')

        self.uiStatusLabel: ttk.Label = self.builder.get_object('uiStatusLabel')
        self.uiLabelPLY: ttk.Label = self.builder.get_object('uiLabelPLY')
        self.uiLabelTXT: ttk.Label = self.builder.get_object('uiLabelTXT')

        self.uiProgressBar: ttk.Progressbar = self.builder.get_object('uiProgressBar')

        self.uiProgressBar.configure(
            orient='horizontal', mode='indeterminate'
        )

        self.uiCheckboxTXT_Checked = tk.BooleanVar()
        self.uiCheckboxPLY_Checked = tk.BooleanVar()
        self.uiCheckbox45Deg_Checked = tk.BooleanVar()

        self.uiStatusLabel_Text = tk.StringVar(value='Choose PLY and TXT files..')
        self.uiLabelPLY_Text = tk.StringVar(value='Not selected..')
        self.uiLabelTXT_Text = tk.StringVar(value='Not selected..')

        self.uiStatusLabel.configure(
            textvariable=self.uiStatusLabel_Text
        )
        self.uiLabelTXT.configure(
            textvariable=self.uiLabelTXT_Text
        )
        self.uiLabelPLY.configure(
            textvariable=self.uiLabelPLY_Text
        )

        self.uiCheckbox45Deg.configure(
            variable=self.uiCheckbox45Deg_Checked
        )
        self.uiCheckboxTXT.configure(
            state=DISABLED,
            variable=self.uiCheckboxTXT_Checked
        )
        self.uiCheckboxPLY.configure(
            state=DISABLED,
            variable=self.uiCheckboxPLY_Checked
        )

        self.uiButtonOpenOk.configure(
            command=self.uiButtonOpenOk_Click
        )
        self.uiButtonCancel.configure(
            command=self.uiButtonCancel_Click
        )

        self.uiButtonOpenTXT.configure(
            command=self.uiButtonOpenTXT_Click
        )
        self.uiButtonOpenPLY.configure(
            command=self.uiButtonOpenPLY_Click
        )

        self.filepath_ply = None
        self.filepath_txt = None
        self.parent: MainWindow = None

        self.is_loading = False

    def show(self, parent):
        self.parent = parent
        self.uiDialog.parent = parent.mainWindow
        self.uiDialog.set_modal(True)
        self.uiDialog.show()
        self.uiDialog.run()

    def uiButtonOpenOk_Click(self):
        if not self.is_loading:
            if self.filepath_ply:
                use_45deg_lines = self.uiCheckbox45Deg_Checked.get()
                self.uiStatusLabel_Text.set('Processing map..')
                self.is_loading = True
                self.uiProgressBar.start()
                self.thread_pool = ThreadPool()
                self.thread_pool.apply_async(main, (self.filepath_ply, self.filepath_txt, use_45deg_lines), callback=self.on_main_callback)

    def on_main_callback(self, map_result):
        self.uiDialog.close()
        self.parent.open_map(map_result, self.filepath_ply)

    def uiButtonOpenTXT_Click(self):
        if not self.is_loading:
            self.filepath_txt = tk.filedialog.askopenfilename(
                title='Open a txt file',
                initialdir='/',
                filetypes=(("txt files", "*.txt"), ("All files", "*.*")))
            if self.filepath_txt:
                self.uiLabelTXT_Text.set(os.path.basename(self.filepath_txt))
                self.uiCheckboxTXT_Checked.set(True)

    def uiButtonOpenPLY_Click(self):
        if not self.is_loading:
            self.filepath_ply = tk.filedialog.askopenfilename(
                title='Open a PLY file',
                initialdir='/',
                filetypes=(("PLY files", "*.ply"), ("All files", "*.*")))
            if self.filepath_ply:
                self.uiLabelPLY_Text.set(os.path.basename(self.filepath_ply))
                self.uiCheckboxPLY_Checked.set(True)

    def uiButtonCancel_Click(self):
        self.uiDialog.close()


class MainWindow:
    def __init__(self, master, maxsize):
        # build ui
        self.mainWindow = ttk.Frame(master)
        master.minsize(800, 600)
        self.master = master

        self.frameMap = ttk.Frame(self.mainWindow)
        self.labelMap = ttk.Label(self.frameMap)
        self.labelMap.configure(anchor='w', font='TkDefaultFont', justify='left', text='Map')
        self.labelMap.pack(expand='false', fill='both', side='top')
        self.framePlot = ttk.Frame(self.frameMap)
        self.framePlot.pack(expand='true', fill='both', side='top')
        self.frameMap.pack(anchor='w', expand='true', fill='both', padx='10', pady='10', side='left')

        figure = plt.Figure(figsize=(10, 10), dpi=150)
        self.fig = figure
        self.ax = figure.add_subplot(111)
        self.canvas = FigureCanvasTkAgg(figure, master=self.framePlot)
        self.canvas.get_tk_widget().pack(fill='both', expand='true')

        self.list_points = []

        self.press = False
        self.move = False

        self.c1 = self.canvas.mpl_connect('button_press_event', self.onpress)
        self.c2 = self.canvas.mpl_connect('button_release_event', self.onrelease)
        self.c3 = self.canvas.mpl_connect('motion_notify_event', self.onmove)

        self.frameFile = ttk.Frame(self.mainWindow)
        self.labelFile = ttk.Label(self.frameFile)
        self.labelFile.configure(anchor='w', compound='top', justify='left', text='File')
        self.labelFile.pack(anchor='s', expand='false', fill='both', side='top')

        self.btnOpenPLY = ttk.Button(self.frameFile, command=self.open_ply_txt, width=20)
        self.btnOpenPLY.configure(text='Load from PLY and TXT..')
        self.btnOpenPLY.pack(anchor='s', expand='false', side='top')

        self.btnOpenHDF5 = ttk.Button(self.frameFile, command=self.open_hdf5, width=20)
        self.btnOpenHDF5.configure(text='Load from HDF5..')
        self.btnOpenHDF5.pack(anchor='s', expand='false', side='top')

        self.btnSaveHDF5 = ttk.Button(self.frameFile, command=self.save_hdf5, width=20)
        self.btnSaveHDF5.configure(text='Save to HDF5..')
        self.btnSaveHDF5.pack(anchor='s', expand='false', side='top')

        self.btnClearData = ttk.Button(self.frameFile, command=self.clear_data, width=20)
        self.btnClearData.configure(text='Clear all')
        self.btnClearData.pack(anchor='s', expand='false', side='top')

        self.frameFile.configure(height='200', width='200')
        self.frameFile.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')

        self.frameLayers = ttk.Frame(self.mainWindow)

        self.labelLayers = ttk.Label(self.frameLayers, justify=LEFT)
        self.labelLayers.configure(anchor='w', text='Layers')
        self.labelLayers.pack(anchor='w', expand='true', fill='x', side='top')

        self.chkWalls = tk.BooleanVar()
        self.chkPointCloud = tk.BooleanVar()
        self.chkDronePath = tk.BooleanVar()
        self.chkPOI = tk.BooleanVar()

        self.chkWalls.set(True)
        self.chkPointCloud.set(True)
        self.chkDronePath.set(True)
        self.chkPOI.set(True)

        self.checkbtnWalls = ttk.Checkbutton(self.frameLayers, variable=self.chkWalls,
                                             command=self.layer_state_change)
        self.checkbtnWalls.configure(text='Walls')
        self.checkbtnWalls.pack(anchor='s', expand='true', fill='both', side='top')
        self.checkbtnPointCloud = ttk.Checkbutton(self.frameLayers, variable=self.chkPointCloud,
                                                  command=self.layer_state_change)
        self.checkbtnPointCloud.configure(text='Point Cloud')
        self.checkbtnPointCloud.pack(anchor='s', expand='true', fill='both', side='top')
        self.checkbtnDronePath = ttk.Checkbutton(self.frameLayers, variable=self.chkDronePath,
                                                 command=self.layer_state_change)
        self.checkbtnDronePath.configure(text='Drone path')
        self.checkbtnDronePath.pack(anchor='s', expand='true', fill='both', side='top')
        self.checkbtnPOI = ttk.Checkbutton(self.frameLayers, variable=self.chkPOI,
                                             command=self.layer_state_change)
        self.checkbtnPOI.configure(text='Points of interest')
        self.checkbtnPOI.pack(anchor='s', expand='true', fill='both', side='top')

        self.frameLayers.pack(anchor='sw', expand='true', fill='x', padx='10', pady='20', side='top')


        self.frameTools = ttk.Frame(self.mainWindow)
        self.labelTools = ttk.Label(self.frameTools)
        self.labelTools.configure(anchor='w', justify='left', text='Tools')
        self.labelTools.pack(anchor='s', expand='false', fill='both', side='top')

        self.radioButtonMainMode_Value = tk.StringVar()
        self.radiobtnViewer = ttk.Radiobutton(self.frameTools, variable=self.radioButtonMainMode_Value, value=TOOL_VIEWER,
                                              command=self.onMainModeChange)
        self.radiobtnViewer.configure(text='Viewer')
        self.radiobtnViewer.pack(anchor='s', expand='false', fill='both', side='top')

        self.radiobtnEditMap = ttk.Radiobutton(self.frameTools, variable=self.radioButtonMainMode_Value, value=TOOL_LINES,
                                               command=self.onMainModeChange)
        self.radiobtnEditMap.configure(text='Edit Map')
        self.radiobtnEditMap.pack(anchor='s', expand='false', fill='both', side='top')

        self.radiobtnEditPoI = ttk.Radiobutton(self.frameTools, variable=self.radioButtonMainMode_Value, value=TOOL_POI,
                                               command=self.onMainModeChange)
        self.radiobtnEditPoI.configure(text='Edit Points of Interest')
        self.radiobtnEditPoI.pack(anchor='s', expand='false', fill='both', side='top')
        self.radioButtonMainMode_Value.set(TOOL_VIEWER)

        self.frameTools.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')

        self.framePoIProperties = ttk.Frame(self.mainWindow)
        self.labelPoIProperties = ttk.Label(self.framePoIProperties)
        self.labelPoIProperties.configure(anchor='w', justify='left', text='Properties - Points of Interest')
        self.labelPoIProperties.pack(anchor='s', expand='false', fill='both', side='top')
        self.radiobtnPoIProperties = tk.StringVar()
        self.radiobtnPropertiesTarget = ttk.Radiobutton(self.framePoIProperties, variable=self.radiobtnPoIProperties,
                                                        value=SUB_TOOL_POI_TARGET,
                                                        command=self.onSubModeChangePOI)
        self.radiobtnPropertiesTarget.configure(text='Target')
        self.radiobtnPropertiesTarget.pack(anchor='s', expand='false', fill='both', side='top')

        self.radiobtnPropertiesDrone = ttk.Radiobutton(self.framePoIProperties, variable=self.radiobtnPoIProperties,
                                                       value=SUB_TOOL_POI_DRONE,
                                                       command=self.onSubModeChangePOI)
        self.radiobtnPropertiesDrone.configure(text='Drone')
        self.radiobtnPropertiesDrone.pack(anchor='s', expand='false', fill='both', side='top')

        self.radiobtnPropertiesStation = ttk.Radiobutton(self.framePoIProperties, variable=self.radiobtnPoIProperties,
                                                         value=SUB_TOOL_POI_STATION,
                                                         command=self.onSubModeChangePOI)
        self.radiobtnPropertiesStation.configure(text='Station')
        self.radiobtnPropertiesStation.pack(anchor='s', expand='false', fill='both', side='top')

        self.labelPOI = tk.Label(
            self.framePoIProperties,
            text='Add with right click. \nRemove with left click.'
        )
        self.labelPOI.pack(anchor='s', fill='x', side='top')

        self.framePoIProperties.configure(height='200', width='200')
        self.framePoIProperties.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')
        self.radiobtnPoIProperties.set(SUB_TOOL_POI_TARGET)
        self.framePoIProperties.pack_forget()

        self.frameEditMapProperties = ttk.Frame(self.mainWindow)
        self.labelEditMapProperties = ttk.Label(self.frameEditMapProperties)
        self.labelEditMapProperties.configure(anchor='w', justify='left', text='Properties - Edit map')
        self.labelEditMapProperties.pack(anchor='s', expand='false', fill='both', side='top')

        self.radioButtonSubModeEditMap_Value = tk.StringVar()
        self.radiobtnPropertiesDrawLine = ttk.Radiobutton(self.frameEditMapProperties,
                                                          variable=self.radioButtonSubModeEditMap_Value,
                                                          value=SUB_TOOL_LINES_ADD,
                                                          command=self.onSubModeChangeEditMap)
        self.radiobtnPropertiesDrawLine.configure(text='Add line')
        self.radiobtnPropertiesDrawLine.pack(anchor='s', expand='false', fill='both', side='top')
        self.radiobtnPropertiesDeleteLine = ttk.Radiobutton(self.frameEditMapProperties,
                                                            variable=self.radioButtonSubModeEditMap_Value,
                                                            value=SUB_TOOL_LINES_MODIFY,
                                                            command=self.onSubModeChangeEditMap)
        self.radiobtnPropertiesDeleteLine.configure(text='Edit existing line')
        self.radiobtnPropertiesDeleteLine.pack(anchor='s', expand='false', fill='both', side='top')

        self.labelEditMap_Value = tk.StringVar(
            value=UI_LABEL_MAP_ADD
        )
        self.labelEditMap = tk.Label(
            self.frameEditMapProperties,
            textvariable=self.labelEditMap_Value,
            justify=LEFT
        )
        self.labelEditMap.pack(anchor='w', fill='x', side='top')
        self.radioButtonSubModeEditMap_Value.set(SUB_TOOL_LINES_ADD)


        self.buttonEditMapApply = ttk.Button(self.frameEditMapProperties, command=self.buttonEditMapApply_Click,
                                             width=20, text='Apply Changes')
        self.buttonEditMapApply.pack(anchor='s', expand='false', side='top')

        self.buttonEditMapCancel = ttk.Button(self.frameEditMapProperties, command=self.buttonEditMapCancel_Click,
                                              width=20, text='Cancel Changes')
        self.buttonEditMapCancel.pack(anchor='s', expand='false', side='top')

        self.buttonEditMapDelete = ttk.Button(self.frameEditMapProperties, command=self.buttonEditMapDelete_Click,
                                             width=20, text='Delete Line')
        self.buttonEditMapDelete.pack(anchor='s', expand='false', side='top')


        self.frameEditMapProperties.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')
        self.frameEditMapProperties.pack_forget()

        self.mainWindow.pack(anchor='w', expand='true', fill='both', side='top')

        # Main widget
        self.mainwindow = self.mainWindow

        master.state('zoomed')

        master.geometry(f'{maxsize[0]}x{maxsize[1]}+0+0')
        self.canvas.get_tk_widget().pack_forget()

        self.frameLayers.pack_forget()
        self.frameTools.pack_forget()
        self.frameEditMapProperties.pack_forget()
        self.framePoIProperties.pack_forget()

        # init other params
        self.map = None
        self.hdf5_opened = False
        self.current_edit_line = None

        self.plt_poi_by_point: Dict[Point, object] = {}
        self.plt_lines_by_idx: Dict[int, object] = {}

    def run(self):
        self.mainwindow.mainloop()

    def callback(self, event=None):
        pass

    def open_ply_txt(self):
        dialog = DialogLoader()
        dialog.show(self)

    def open_map(self, map, filename, is_rotate=True):

        self.master.title(os.path.basename(filename))

        if self.hdf5_opened:
            self.clear_data()
            self.hdf5_opened = False

        self.plt_poi_by_point: Dict[Point, object] = {}
        self.plt_lines_by_idx: Dict[int, object] = {}

        self.map = map
        if is_rotate:
            self.map.rotate_and_scale_data()
        self.plot_using_map_data()

        self.canvas.get_tk_widget().pack()
        self.frameLayers.pack()
        self.frameTools.pack()
        self.radioButtonMainMode_Value.set(TOOL_VIEWER)
        self.chkWalls.set(True)
        self.chkPointCloud.set(True)
        self.chkDronePath.set(True)
        self.chkPOI.set(True)



    def open_hdf5(self):
        try:
            filename = tk.filedialog.askopenfilename(
                title='Open a HDF5 file',
                filetypes=(("HDF5 files", "*.h5"), ("All files", "*.*")))

            if filename:
                with open('./args.json') as json_file:
                    args = DictToObj(**json.load(json_file))

                self.hdf5_opened = True
                self.map = Map(args=args, hf5_path=filename)
                self.open_map(self.map, filename, is_rotate=False)

        except Exception as e:
            logger.exception(e)

    def clear_data(self):
        self.map = None
        self.ax.clear()
        self.canvas.draw()

        self.frameLayers.pack_forget()
        self.frameTools.pack_forget()
        self.frameEditMapProperties.pack_forget()
        self.framePoIProperties.pack_forget()

    def save_hdf5(self):
        try:
            file_path = tk.filedialog.asksaveasfilename(
                defaultextension='',
                initialfile='',
                title='Save HDF5',
                filetypes=(("HDF5 files", "*.h5"), ("All files", "*.*")))
            if file_path:
                if not file_path.endswith(".h5"):
                    file_path += '.h5'
                self.map.save_to_h5(file_path)
                self.master.title(os.path.basename(file_path))
        except Exception as e:
            logger.exception(e)

    # ---PLOT FUNCTIONS---
    def plot_using_map_data(self):
        if self.map is not None:
            self.ax.clear()
            if self.chkWalls.get():
                # line_start_points, line_end_points = self.map.get_line_coordinates_in_meters()
                for idx in range(len(self.map.line_start_points_in_meters)):
                    plt_plot, = self.ax.plot(
                        [
                            self.map.line_start_points_in_meters[idx, 0],
                            self.map.line_end_points_in_meters[idx, 0]
                        ],
                        [
                            self.map.line_start_points_in_meters[idx, 1],
                            self.map.line_end_points_in_meters[idx, 1]
                        ],
                        c='darkred', linewidth=3.0)
                    self.plt_lines_by_idx[idx] = plt_plot

            if self.chkPointCloud.get() and self.map.point_cloud is not None:
                x = map(lambda point: point.x, self.map.point_cloud)
                y = map(lambda point: point.y, self.map.point_cloud)
                self.ax.scatter(list(x), list(y), s=1.0, c='gray')

            if self.chkDronePath.get():
                x = map(lambda point: point.x, self.map.path_points)
                y = map(lambda point: point.y, self.map.path_points)
                self.ax.plot(list(x), list(y), c='blue', linewidth=2)

            if self.chkPOI.get():
                # TODO fix ineffective POI redrawing (canvas.draw() is called after each icon is added to ax)
                if len(self.map.poi_target) != 0:
                    for point in self.map.poi_target:
                        self.plot_poi(point, SUB_TOOL_POI_TARGET)
                if len(self.map.poi_station) != 0:
                    for point in self.map.poi_station:
                        self.plot_poi(point, SUB_TOOL_POI_STATION)
                if len(self.map.poi_drone) != 0:
                    for point in self.map.poi_drone:
                        self.plot_poi(point, SUB_TOOL_POI_DRONE)
            self.ax.set_aspect('equal')
            self.ax.invert_yaxis()
            self.ax.invert_xaxis()
            self.ax.get_xaxis().set_visible(False)
            self.ax.get_yaxis().set_visible(False)

            self.canvas.draw()

    def plot_poi(self, point, poi_type):
        # TODO change pictures
        if poi_type == SUB_TOOL_POI_TARGET:
            img = mpimg.imread('./media/icons8-target.png')
        elif poi_type == SUB_TOOL_POI_STATION:
            img = mpimg.imread('./media/icons8-home.png')
        elif poi_type == SUB_TOOL_POI_DRONE:
            img = mpimg.imread('./media/icons8-drone.png')
        img = OffsetImage(img, zoom=0.3)
        img = AnnotationBbox(img, (point.x, point.y), frameon=False)

        self.plt_poi_by_point[point] = img
        self.ax.add_artist(img)
        self.canvas.draw()

    # ---MOUSE ACTION FUNCTIONS---

    def onpress(self, event):
        tools_state = self.radioButtonMainMode_Value.get()
        point = Point(event.lastevent.xdata, event.lastevent.ydata)
        if self.map is not None:
            if tools_state == TOOL_LINES:
                editMap_state = self.radioButtonSubModeEditMap_Value.get()
                if editMap_state == SUB_TOOL_LINES_ADD:
                    # TODO handle new line drawing

                    if len(self.list_points) < 2:
                        self.list_points.append(PltDraggablePoint(self, x=point.x, y=point.y))
                        self.canvas.draw()

                    if event.button == MouseButton.RIGHT:
                        self.remove_line_drawer()

                if editMap_state == SUB_TOOL_LINES_MODIFY:
                    if len(self.list_points) == 0: # currently no line in edit mode
                        for idx, plt_plot in self.plt_lines_by_idx.items():
                            contains, attrd = plt_plot.contains(event)
                            if contains:
                                self.current_edit_line = [
                                    self.map.line_start_points_in_meters[idx],
                                    self.map.line_end_points_in_meters[idx]
                                ]
                                self.map.line_start_points_in_meters = np.delete(self.map.line_start_points_in_meters, idx, axis=0)
                                self.map.line_end_points_in_meters = np.delete(self.map.line_end_points_in_meters, idx, axis=0)
                                self.plot_using_map_data()

                                self.list_points.append(PltDraggablePoint(self, x=self.current_edit_line[0][0], y=self.current_edit_line[0][1]))
                                self.list_points.append(PltDraggablePoint(self, x=self.current_edit_line[1][0], y=self.current_edit_line[1][1]))
                                self.canvas.draw()
                                break

            elif tools_state == TOOL_POI:
                if event.button == MouseButton.LEFT:
                    editPoI_state = self.radiobtnPoIProperties.get()
                    if editPoI_state == SUB_TOOL_POI_TARGET:
                        self.map.poi_target.append(point)
                    elif editPoI_state == SUB_TOOL_POI_STATION:
                        self.map.poi_station.append(point)
                    elif editPoI_state == SUB_TOOL_POI_DRONE:
                        self.map.poi_drone.append(point)
                    self.plot_poi(point, editPoI_state)

                elif event.button == MouseButton.RIGHT:
                    for point, img in self.plt_poi_by_point.items():
                        contains, attrd = img.patch.contains(event)
                        if contains:
                            img.remove()
                            self.canvas.draw()
                            del self.plt_poi_by_point[point]
                            if point in self.map.poi_target:
                                self.map.poi_target.remove(point)
                            elif point in self.map.poi_station:
                                self.map.poi_station.remove(point)
                            elif point in self.map.poi_drone:
                                self.map.poi_drone.remove(point)
                            break

            print('click', event)
            pass

    def onmove(self, event):
        #print('onmove', event)
        if self.press:
            self.move = True

    def onrelease(self, event):
        #print('onrelease', event)
        if self.press and not self.move:
            self.onclick(event)
        self.press = False
        self.move = False

    def layer_state_change(self):
        self.plot_using_map_data()

    def remove_line_drawer(self):
        for each in self.list_points:
            each.on_release(None)
            each.destroy()
        self.canvas.draw()
        self.list_points = []

    def onMainModeChange(self):
        val = self.radioButtonMainMode_Value.get()
        if val == TOOL_VIEWER:
            self.frameEditMapProperties.pack_forget()
            self.framePoIProperties.pack_forget()
            self.plot_using_map_data()
        elif val == TOOL_POI:
            self.frameEditMapProperties.pack_forget()
            self.framePoIProperties.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')
            self.chkPOI.set(True)
            self.plot_using_map_data()

        if val == TOOL_LINES:
            self.frameEditMapProperties.pack(anchor='s', expand='false', fill='both', padx='10', pady='5', side='top')
            self.framePoIProperties.pack_forget()
            self.chkPOI.set(False)
        else:
            self.remove_line_drawer()

    def onSubModeChangePOI(self):
        pass

    def onSubModeChangeEditMap(self):
        val = self.radioButtonSubModeEditMap_Value.get()
        if val == SUB_TOOL_LINES_MODIFY:
            self.labelEditMap_Value.set(UI_LABEL_MAP_MODIFY)
        elif val == SUB_TOOL_LINES_ADD:
            self.labelEditMap_Value.set(UI_LABEL_MAP_ADD)

    def buttonEditMapDelete_Click(self):
        self.remove_line_drawer()

    def buttonEditMapCancel_Click(self):
        self.remove_line_drawer()

        if self.current_edit_line:
            self.map.line_start_points_in_meters = np.concatenate(
                (self.map.line_start_points_in_meters, np.array([self.current_edit_line[0]])), axis=0)
            self.map.line_end_points_in_meters = np.concatenate(
                (self.map.line_end_points_in_meters, np.array([self.current_edit_line[1]])), axis=0)
            self.plot_using_map_data()

    def buttonEditMapApply_Click(self):

        start = self.list_points[0]
        end = self.list_points[1]
        self.map.line_start_points_in_meters = np.concatenate((self.map.line_start_points_in_meters, np.array([[start.x, start.y, 1.0]])), axis=0)
        self.map.line_end_points_in_meters = np.concatenate((self.map.line_end_points_in_meters, np.array([[end.x, end.y, 1.0]])), axis=0)
        self.plot_using_map_data()
        self.remove_line_drawer()


if __name__ == '__main__':
    root = tk.Tk()
    root.title("plyedit 0.2")
    maxsize = root.maxsize()
    app = MainWindow(root, maxsize)
    app.run()


