import numpy as np
import random
from skimage.transform import hough_line, hough_line_peaks
import matplotlib.pyplot as plt
import cv2  # pip install opencv-python
from scipy import ndimage
import copy
from modules.lines_utils import *

def generate_colours(n):
    get_colors = lambda n: list(map(lambda i: "#" + "%06x" % random.randint(0, 0xFFFFFF), range(n)))
    return get_colors(n)


def hex_to_rgb(h):
    h = h.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


def normalize(y, min=-1, max=1):
    min_val = np.min(y)
    max_val = np.max(y)
    y[:] = (y - min_val) / (max_val - min_val)
    f = lambda x: x * (max_val - min_val) + min_val
    if min != 0 or max != 1:
        y[:] = y * (max - min) + min
        f = lambda x: (x - min) / (max - min) * (max_val - min_val) + min_val
    return y, f

def standardize(list):
    mean = np.mean(list)
    std = np.std(list)

    list = (list - mean) / std
    f = lambda x: x * std + mean
    return list, f


def lines_to_features(args, lines):
    features = []
    unnormalized = []
    if args.use_midpoints:
        midpoints_x = []
        midpoints_y = []
        for line in lines:
            midpoint = line.midpoint
            midpoints_x.append(midpoint.x)
            midpoints_y.append(midpoint.y)
        midpoints_x = np.array(midpoints_x)
        midpoints_y = np.array(midpoints_y)

        if args.is_normalize:
            if args.normalize_type.lower() == 'minmax':
                midpoints_x, unnorm_midpoints_x = normalize(midpoints_x)
                midpoints_y, unnorm_midpoints_y = normalize(midpoints_y)
            else:
                midpoints_x, unnorm_midpoints_x = standardize(midpoints_x)
                midpoints_y, unnorm_midpoints_y = standardize(midpoints_y)
        features.extend([midpoints_x, midpoints_y])
        unnormalized.extend([unnorm_midpoints_x, unnorm_midpoints_y])

        if args.use_line_length:
            length = []
            for line in lines:
                length.append(line.length)
            length = np.array(length)
            if args.is_normalize:
                if args.normalize_type.lower() == 'minmax':
                    length, unnorm_length_x = normalize(length)
                else:
                    length, unnorm_length_x = standardize(length)
            features.extend([length])
            unnormalized.extend([unnorm_length_x])

        if args.angle_as_sin_cos:
            sin_vals = []
            cos_vals = []
            for line in lines:
                sin, cos = line.angle_as_sin_cos
                sin_vals.append(sin)
                cos_vals.append(cos)
            sin_vals = np.array(sin_vals)
            cos_vals = np.array(cos_vals)
            features.extend([sin_vals, cos_vals])
        else:
            angle = []
            for line in lines:
                angle.append(line.angle)
            angle = np.array(angle)
            features.extend([angle])

    else:
        start_x = []
        start_y = []
        end_x = []
        end_y = []
        for line in lines:
            start_x.append(line.start_point.x)
            start_y.append(line.start_point.y)
            end_x.append(line.end_point.x)
            end_y.append(line.end_point.y)
        start_x = np.array(start_x)
        start_y = np.array(start_y)
        end_x = np.array(end_x)
        end_y = np.array(end_y)

        if args.is_normalize:
            if args.normalize_type.lower() == 'minmax':
                start_x, unnorm_start_x = normalize(start_x)
                start_y, unnorm_start_y = normalize(start_y)
                end_x, unnorm_end_x = normalize(end_x)
                end_y, unnorm_end_y = normalize(end_y)
            else:
                start_x, unnorm_start_x = standardize(start_x)
                start_y, unnorm_start_y = standardize(start_y)
                end_x, unnorm_end_x = standardize(end_x)
                end_y, unnorm_end_y = standardize(end_y)

        features.extend([start_x, start_y, end_x, end_y])
        unnormalized.extend([unnorm_start_x, unnorm_start_y, unnorm_end_x, unnorm_end_y])

    features = np.stack(features, axis=1)
    return features, unnormalized


def get_endpoints(args, image, x, y, angle, length=None):
    map_binary = np.array(np.where(image == (255, 255, 255), 0, 255), dtype=np.uint8)[:,:,0]
    poly_width = args.metric_line_width
    img_x = image.shape[1]
    img_y = image.shape[0]
    if angle > np.pi / 2:
        angle -= np.pi
    elif angle < -np.pi / 2:
        angle += np.pi
    if length is None:
        if abs(angle) > np.pi / 4:
            y_start = 0
            y_end = img_y
            x_start = x + (y_start - y) / np.tan(angle)
            x_end = x + (y_end - y) / np.tan(angle)
        else:
            x_start = 0
            x_end = img_x
            y_start = y + np.tan(angle) * (x_start - x)
            y_end = y + np.tan(angle) * (x_end - x)
    else:
        half_len = length / 2
        x_start = x - half_len * np.cos(angle)
        x_end = x + half_len * np.cos(angle)
        y_start = y - half_len * np.sin(angle)
        y_end = y + half_len * np.sin(angle)

    line = Line(Point(x_start, y_start), Point(x_end, y_end))
    # polygon points
    p0 = (x_start + poly_width * np.cos(angle + np.pi/2), y_start + poly_width * np.sin(angle + np.pi/2))
    p1 = (x_start - poly_width * np.cos(angle + np.pi/2), y_start - poly_width * np.sin(angle + np.pi/2))
    p2 = (x_end - poly_width * np.cos(angle + np.pi/2), y_end - poly_width * np.sin(angle + np.pi/2))
    p3 = (x_end + poly_width * np.cos(angle + np.pi/2), y_end + poly_width * np.sin(angle + np.pi/2))
    coords = [p0, p1, p2, p3, p0]
    mask = np.zeros(image.shape[:2], dtype=bool)
    fill_polygon(copy.copy(coords), mask)
    masked_map = np.logical_and(mask, map_binary)

    if length is None:
        # find line endpoints
        if np.max(masked_map):
            x_y_deltas = [i.max()-i.min() for i in np.where(masked_map)]
            if np.argmax(x_y_deltas) == 0:
                y_mask = np.where(masked_map)[0]
                start_y = np.min(y_mask)
                end_y = np.max(y_mask)
                start_x = x + (start_y - y) / np.tan(angle)
                end_x = x + (end_y - y) / np.tan(angle)
            else:
                x_mask = np.where(masked_map)[1]
                start_x = np.min(x_mask)
                end_x = np.max(x_mask)
                start_y = y + np.tan(angle) * (start_x - x)
                end_y = y + np.tan(angle) * (end_x - x)

            line = Line(Point(start_x, start_y), Point(end_x, end_y))
        else:
            line = None

    if line is not None:
        if line.length < args.min_line_len:
            line = None
            line_pixels = 0
        else:
            line_pixels = np.sum(masked_map)

    return line, line_pixels


def features_to_lines(args, image, features, unnormalize):
    # unnormalize features
    for idx in range(len(unnormalize)):
        features[:, idx] = unnormalize[idx](features[:, idx])
    lines = []
    lines_pixels = []
    if args.use_midpoints:
        x_mid = features[:, 0]
        y_mid = features[:, 1]
        if args.use_line_length:
            length = features[:, 2]
            if args.angle_as_sin_cos:
                angle = np.arctan2(features[:, 3], features[:, 4])
            else:
                angle = features[:, 3]
        else:
            if args.angle_as_sin_cos:
                angle = np.arctan2(features[:, 2], features[:, 3])
            else:
                angle = features[:, 2]

        for idx in range(len(features)):
            length_val = length[idx] if args.use_line_length else None
            line, line_pixels = get_endpoints(args, image, x_mid[idx], y_mid[idx], angle[idx], length_val)
            if line is not None:
                lines.append(line)
                lines_pixels.append(line_pixels)

    else:
        x_start = features[:, 0]
        y_start = features[:, 1]
        x_end = features[:, 2]
        y_end = features[:, 3]
        for idx in range(len(features)):
            line = Line((x_start[idx], y_start[idx]), (x_end[idx], y_end[idx]))
            lines.append(line)

    return lines, lines_pixels


def fill_polygon(polygon, canvas):
    if not len(polygon):
        return
    for idx in range(len(polygon)):
        polygon[idx] = polygon[idx][::-1]
    min_y = min(y for y,x in polygon)
    max_y = max(y for y,x in polygon)
    polygon = [(float(y),float(x)) for y,x in polygon]
    if min_y < canvas.shape[0]:
        min_y = 0
    if max_y >= canvas.shape[0]:
        max_y = canvas.shape[0]

    for y in range(int(min_y), int(max_y)):
        nodes = []
        j = -1
        for i,p in enumerate(polygon):
            pj = polygon[j]
            if p[0] < y and pj[0] >= y or pj[0] < y and p[0] >= y:
                dy = pj[0] - p[0]
                if dy:
                    nodes.append( (p[1] + (y-p[0])/(pj[0]-p[0])*(pj[1]-p[1])) )
                elif p[0] == y:
                    nodes.append(p[1])
            j = i
        nodes.sort()
        for n,nn in zip(nodes[::2],nodes[1::2]):
            nn += 1
            canvas[y, int(n):int(nn)] = True
