import cv2  # pip install opencv-python
import numpy as np
import os
import copy
from scipy import ndimage

from skimage.transform import hough_line, hough_line_peaks
from skimage.feature import canny
from skimage.draw import line
from skimage import data

import matplotlib.pyplot as plt
from matplotlib import cm

from modules.ply_reader import *
from modules.lines_utils import *


class HoughTransform:
    @staticmethod
    def rotate_image(image, show=False):
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 360, endpoint=False)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image = np.array(np.where(image == 255, 0, 255), dtype=np.uint8)
        h, theta, d = hough_line(image, theta=tested_angles)
        h, theta, d = hough_line_peaks(h, theta, d)
        rotate_angle = theta[np.argmax(h)]
        image_rotated = ndimage.rotate(image, 180*rotate_angle/np.pi)
        if show:
            fig, ax = plt.subplots(2, 1, figsize=(10, 10))
            ax[0].imshow(image)
            ax[1].imshow(image_rotated)
            plt.show(block=False)
        return image_rotated


    @staticmethod
    def detect_lines(img, args):
        if img.ndim == 2:
            blackAndWhiteImage = img
        else:
            # Convert the image to gray-scale
            grayImage = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            (thresh, blackAndWhiteImage) = cv2.threshold(grayImage, 127, 255, cv2.THRESH_BINARY)

        # Find the edges in the image using canny detector
        edges = cv2.Canny(blackAndWhiteImage,
                          threshold1=args.canny_threshold1,
                          threshold2=args.canny_threshold1,
                          apertureSize=args.canny_aperatureSize,
                          L2gradient=args.canny_L2grad)
        # Detect points that form a line
        if args.is_probabilistic_hough:
            lines = cv2.HoughLinesP(edges,
                                    rho=args.hough_rho,
                                    theta=args.hough_theta,
                                    threshold=args.hough_threshold,
                                    minLineLength=args.hough_minLineLength,
                                    maxLineGap=args.hough_maxLineGap)
            lines_obj = []
            for line in lines:
                lines_obj.append(Line((line[0, 0], line[0, 1]), (line[0, 2], line[0, 3])))
        else:
            lines = cv2.HoughLines(edges,
                                   rho=args.hough_rho,
                                   theta=args.hough_theta,
                                   threshold=args.hough_threshold)
            lines_obj = []
            for line in lines:
                for rho, theta in line:
                    a = np.cos(theta)
                    b = np.sin(theta)
                    x0 = a * rho
                    y0 = b * rho
                    x1 = int(x0 + 1000 * (-b))
                    y1 = int(y0 + 1000 * (a))
                    x2 = int(x0 - 1000 * (-b))
                    y2 = int(y0 - 1000 * (a))
                    lines_obj.append(Line((x1, y1), (x2, y2)))

            # Hough lines with scikit
            # lines_obj = []
            # tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 360, endpoint=False)
            # h, theta, d = hough_line(edges,
            #                         theta=tested_angles)
            # for _, angle, dist in zip(*hough_line_peaks(h, theta, d)):
            #     (x0, y0) = dist * np.array([np.cos(angle), np.sin(angle)])
            #     a = np.cos(angle)
            #     b = np.sin(angle)
            #     x1 = int(x0 + 1000 * (-b))
            #     y1 = int(y0 + 1000 * (a))
            #     x2 = int(x0 - 1000 * (-b))
            #     y2 = int(y0 - 1000 * (a))
            #     lines_obj.append(Line((x1, y1), (x2, y2)))

        return lines_obj

    @staticmethod
    def detect_blobs(img, args, show=False):
        # Good visualisation of parameters:
        # https://learnopencv.com/blob-detection-using-opencv-python-c/

        # Setup SimpleBlobDetector parameters
        params = cv2.SimpleBlobDetector_Params()

        # Change thresholds
        params.thresholdStep = args.blobs_thresholdStep
        params.minThreshold = args.blobs_minThreshold
        params.maxThreshold = args.blobs_maxThreshold

        # Filter by color
        params.filterByColor = False

        params.minDistBetweenBlobs = args.blobs_minDistBetweenBlobs

        # Filter by Area.
        params.filterByArea = True
        params.minArea = args.blobs_minArea
        params.maxArea = args.blobs_maxArea

        # Filter by Circularity
        params.filterByCircularity = True
        params.minCircularity = args.blobs_minCircularity
        # params.maxCircularity = 1

        # Filter by Convexity
        params.filterByConvexity = True
        params.minConvexity = args.blobs_minConvexity
        # params.maxConvexity = 1

        # Filter by Inertia
        params.filterByInertia = True
        params.minInertiaRatio = args.blobs_minInertiaRatio

        # Create a detector with the parameters
        detector = cv2.SimpleBlobDetector_create(params)

        # Detect blobs.
        keypoints = detector.detect(img)

        # Draw detected blobs as red circles.
        # cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS ensures the size of the circle corresponds to the size of blob
        im_with_keypoints = cv2.drawKeypoints(img, keypoints, np.array([]), (0, 0, 255),
                                              cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)

        img_without_blobs = img
        for keypoint in keypoints:
            img_without_blobs = cv2.circle(img_without_blobs,
                                           center=(int(keypoint.pt[0]), int(keypoint.pt[1])),
                                           radius=int(keypoint.size/2),
                                           color=(0, 0, 0),
                                           thickness=-1)

        if show:
            cv2.imshow("Keypoints", im_with_keypoints)
            cv2.waitKey(0)

            cv2.imshow("Keypoints colored out", img_without_blobs)
            cv2.waitKey(0)

            cv2.destroyAllWindows()

        return img_without_blobs

    @staticmethod
    def plot(img, lines, show=True, points=None, empty=False):
        if empty:
            image = np.zeros_like(img)
        else:
            image = copy.copy(img)

        if img.ndim == 2:
            image = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        for line in lines:
            x1 = line.start_point.x
            y1 = line.start_point.y
            x2 = line.end_point.x
            y2 = line.end_point.y
            cv2.line(image, (int(x1), int(y1)), (int(x2), int(y2)), (255, 0, 0), 3)

        if points != None:
            for point in points:
                x, y = point
                cv2.circle(image, (x, y), radius=1, color=(0, 0, 255), thickness=-1)

        # Show result
        if show:
            cv2.imshow("Hough transform result image", image)
            cv2.waitKey()  # cv2.imshow won't work without this :)
            cv2.destroyAllWindows()
        return image


if __name__ == '__main__':
    max_slider = 30
    directory = r'C:\Users\aigaa\Documents\GitHub\alarm_pointcloud_backend_2021_Q4\data'

    for file in os.listdir(directory):
        filename = os.fsdecode(file)
        if filename.endswith(".ply"):
            path = os.path.join(directory, filename)
            vertices = PlyReader.read(path)
            img = PlyReader.points_3D_to_image(vertices, show=True)
            img = HoughTransform.detect_blobs(img, size='small', show=True)
            img = HoughTransform.rotate_image(img, show=False)

            # image = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

            # length_threshold = 10
            # distance_threshold = 1.41421356
            # canny_th1 = 50.0
            # canny_th2 = 50.0
            # canny_aperture_size = 3
            # do_merge = True
            # fld = cv2.ximgproc.createFastLineDetector(
            #     length_threshold,
            #     distance_threshold,
            #     canny_th1,
            #     canny_th2,
            #     canny_aperture_size,
            #     do_merge
            # )
            # lines = fld.detect(img)
            # HoughTransform.plot(img, lines, show=True)

            # lines = HoughTransform.detect_lines(img, max_slider)
            # HoughTransform.plot(img, lines, show=True)
        else:
            continue

    # Scikit approach:

    # # Classic straight-line Hough transform
    # # Set a precision of 0.5 degree.
    # tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 360, endpoint=False)
    # img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    #
    # img = np.where(img == 255, 0, 255)
    #
    # h, theta, d = hough_line(img, theta=tested_angles)
    #
    # # Generating figure 1
    # fig, axes = plt.subplots(1, 3, figsize=(15, 6))
    # ax = axes.ravel()
    #
    # ax[0].imshow(img, cmap=cm.gray)
    # ax[0].set_title('Input image')
    # ax[0].set_axis_off()
    #
    # angle_step = 0.5 * np.diff(theta).mean()
    # d_step = 0.5 * np.diff(d).mean()
    # bounds = [np.rad2deg(theta[0] - angle_step),
    #           np.rad2deg(theta[-1] + angle_step),
    #           d[-1] + d_step, d[0] - d_step]
    # ax[1].imshow(np.log(1 + h), extent=bounds, cmap=cm.gray, aspect=1 / 1.5)
    # ax[1].set_title('Hough transform')
    # ax[1].set_xlabel('Angles (degrees)')
    # ax[1].set_ylabel('Distance (pixels)')
    # ax[1].axis('image')
    #
    # ax[2].imshow(img, cmap=cm.gray)
    # ax[2].set_ylim((img.shape[0], 0))
    # ax[2].set_axis_off()
    # ax[2].set_title('Detected lines')
    #
    # for _, angle, dist in zip(*hough_line_peaks(h, theta, d)):
    #     (x0, y0) = dist * np.array([np.cos(angle), np.sin(angle)])
    #     ax[2].axline((x0, y0), slope=np.tan(angle + np.pi / 2))
    #
    # plt.tight_layout()
    # plt.show()


