import numpy as np
import copy
from shapely.geometry.linestring import LineString


class Point():
    def __init__(self, x, y):
        self.x = x
        self.y = y

    @property
    def as_list(self):
        return [float(self.x), float(self.y)]

    @property
    def as_int_list(self):
        return [int(self.x), int(self.y)]

    def distance(self, point):
        return np.sqrt((self.x - point.x)**2 + (self.y - point.y)**2)

    def transform(self, transform):
        new_point = np.matmul(transform, np.array([self.x, self.y, 1]))
        return Point(new_point[0], new_point[1])

    def __str__(self):
        return ','.join([str(it) for it in self.as_int_list])

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))


class Point3D(Point):
    def __init__(self, x, y, z):
        super(Point3D, self).__init__(x, y)
        self.z = z

    @property
    def as_list(self):
        return [float(self.x), float(self.y), float(self.z)]

    @property
    def as_int_list(self):
        return [int(self.x), int(self.y), int(self.z)]

    def distance(self, point):
        return np.sqrt((self.x - point.x)**2 + (self.y - point.y)**2, (self.z - point.z)**2)

    def to_2d(self):
        return Point(self.x, self.y)

    def __str__(self):
        return ','.join([str(it) for it in self.as_int_list])

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))


class Line():
    def __init__(self, p1, p2, angle=None):
        if isinstance(p1, Point):
            self.start_point = p1
        else:
            self.start_point = Point(p1[0], p1[1])
        if isinstance(p2, Point):
            self.end_point = p2
        else:
            self.end_point = Point(p2[0], p2[1])

        # order points so start closer to 0,0
        p1_dist = np.sqrt((self.start_point.x - 0)**2 + (self.start_point.y - 0)**2)
        p2_dist = np.sqrt((self.end_point.x - 0)**2 + (self.end_point.y - 0)**2)
        if p2_dist < p1_dist:
            tmp_p = self.start_point
            self.start_point = self.end_point
            self.end_point = tmp_p

        if angle is None:
            self.angle = np.arctan2(self.end_point.y - self.start_point.y, self.end_point.x - self.start_point.x)
        else:
            self.angle = angle

    def delta_angle(self, angle):
        return np.atan2(np.sin(angle - self.angle), np.cos(angle - self.angle))

    def intersects_line(self, line):
        linestring = self.as_linestring
        return linestring.intersection(line.as_linestring) if linestring.intersects(line.as_linestring) else False

    def transform(self, transform):
        new_start_point = self.start_point.transform(transform)
        new_end_point = self.end_point.transform(transform)
        return Line(new_start_point, new_end_point)

    def extend_line(self, length, extent_end):
        tmp_line = copy.deepcopy(self)
        if extent_end:
            tmp_line.end_point.x = tmp_line.end_point.x + length * np.cos(tmp_line.angle)
            tmp_line.end_point.y = tmp_line.end_point.y + length * np.sin(tmp_line.angle)
        else:
            tmp_line.start_point.x = tmp_line.start_point.x - length * np.cos(tmp_line.angle)
            tmp_line.start_point.y = tmp_line.start_point.y - length * np.sin(tmp_line.angle)
        return tmp_line

    def shorten_line(self, length, cut_end):
        tmp_line = copy.deepcopy(self)
        if length > self.length:
            return None
        if cut_end:
            tmp_line.end_point.x = tmp_line.end_point.x - length * np.cos(tmp_line.angle)
            tmp_line.end_point.y = tmp_line.end_point.y - length * np.sin(tmp_line.angle)
        else:
            tmp_line.start_point.x = tmp_line.start_point.x + length * np.cos(tmp_line.angle)
            tmp_line.start_point.y = tmp_line.start_point.y + length * np.sin(tmp_line.angle)
        return tmp_line

    def rotate_line(self, angle):
        # absolute angle
        tmp_line = copy.deepcopy(self)
        len_from_midpoint = self.midpoint.distance(self.start_point)
        tmp_line.start_point.x = self.midpoint.x - len_from_midpoint * np.cos(angle)
        tmp_line.start_point.y = self.midpoint.y - len_from_midpoint * np.sin(angle)
        tmp_line.end_point.x = self.midpoint.x + len_from_midpoint * np.cos(angle)
        tmp_line.end_point.y = self.midpoint.y + len_from_midpoint * np.sin(angle)
        tmp_line.angle = angle
        return tmp_line

    def side_shift_line(self, distance):
        # absolute angle
        tmp_line = copy.deepcopy(self)
        tmp_line.start_point.x = self.start_point.x + distance * np.cos(self.angle + np.pi/2)
        tmp_line.start_point.y = self.start_point.y + distance * np.sin(self.angle + np.pi/2)
        tmp_line.end_point.x = self.end_point.x + distance * np.cos(self.angle + np.pi/2)
        tmp_line.end_point.y = self.end_point.y + distance * np.sin(self.angle + np.pi/2)
        return tmp_line

    @property
    def as_list(self):
        return [*self.start_point.as_list, *self.end_point.as_list]

    @property
    def midpoint(self):
        return Point(x=(self.start_point.x + self.end_point.x) / 2, y=(self.start_point.y + self.end_point.y) / 2)

    @property
    def length(self):
        return np.sqrt((self.start_point.x - self.end_point.x)**2 + (self.start_point.y - self.end_point.y)**2)

    @property
    def angle_as_sin_cos(self):
        return np.sin(self.angle), np.cos(self.angle)

    @property
    def as_linestring(self):
        return LineString((self.start_point.as_list, self.end_point.as_list))

    def __str__(self):
        return ','.join([str(it) for it in self.as_list])

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))
