import numpy as np
from plyfile import PlyData  # pip install plyfile
import sys
import cv2
import os
import copy
import io
from scipy import ndimage
import json
import h5py
from loguru import logger
from skimage.transform import hough_line, hough_line_peaks
from shapely.geometry.linestring import LineString
from shapely.ops import nearest_points

from modules.lines_utils import Point3D, Point, Line


class Map():
    def __init__(self, args, ply_filepath=None, path_filepath=None, point_cloud=None, path_points=None, hf5_path=None, use_45deg_lines=False):
        self.args = args
        self.point_cloud = None
        self.path_points = None
        self.poi_target = []
        self.poi_station = []
        self.poi_drone = []
        self.map_image = None
        self.raw_image = None
        self.map_final = None
        self.pix_to_meters = None
        self.meters_to_pix = None
        self.lines = []
        self.line_start_points_in_meters = None
        self.line_end_points_in_meters = None
        self.path_points = []
        if ply_filepath is not None:
            self.ply_from_path(ply_filepath)
        if path_filepath is not None:
            self.drone_path_from_path(path_filepath)
        else:
            if point_cloud is not None:
                self.point_cloud = [Point3D(p[0], p[1], p[2]) for p in point_cloud]
            if path_points is not None:
                self.path_points = [Point3D(p[0], p[1], p[2]) for p in path_points]

        args.use_45deg_lines = use_45deg_lines

        if hf5_path is not None:
            self.read_from_h5(hf5_path)
        else:
            self.__init()

    @property
    def is_path_points(self):
        return np.any(self.path_points)

    def ply_from_path(self, ply_filepath):
        with open(ply_filepath, 'rb') as f:
            plydata = PlyData.read(f)

        vertices_x = plydata.elements[0].data['x']
        vertices_y = plydata.elements[0].data['y']
        vertices_z = plydata.elements[0].data['z']
        vertices = np.column_stack((vertices_x, vertices_y, vertices_z))
        self.point_cloud = [Point3D(p[0], p[1], p[2]) for p in vertices]

    def drone_path_from_path(self, path_filepath):
        if os.path.isfile(path_filepath):
            self.path_points = []
            with open(path_filepath, 'rb') as f:
                lines = f.readlines()
                for idx in range(len(lines) - 2):
                    line = lines[idx].decode('utf-8')
                    line = line.split(sep=" ")
                    self.path_points.append(Point3D(float(line[3]), float(line[7]), float(line[11])))

    def __rotate_map(self):
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 360, endpoint=False)
        image = self.map_image[:, :, 0]
        h, theta, d = hough_line(image, theta=tested_angles)
        h, theta, d = hough_line_peaks(h, theta, d)
        rotate_angle = theta[np.argmax(h)] - np.pi / 2
        # rotate map
        self.map_image = ndimage.rotate(self.map_image, 180*rotate_angle/np.pi, reshape=False)
        # self.map_image = np.array(np.where(self.map_image > (100,100,100), (255,255,255), (0,0,0)), dtype=np.uint8)
        # add rotation to transformation matrix
        center_image_pix = Point(*np.array(self.map_image.shape[:2]) / 2)
        self.center_image_pix = copy.deepcopy(center_image_pix)
        center_image_meters = center_image_pix.transform(self.pix_to_meters)
        self.translate_to_image = np.array([
            [1, 0, -center_image_meters.x],
            [0, 1, -center_image_meters.y],
            [0, 0, 1]
        ])
        rotate_angle = -rotate_angle
        self.rotate_matrix = np.array([
            [np.cos(rotate_angle), -np.sin(rotate_angle), 0],
            [np.sin(rotate_angle), np.cos(rotate_angle), 0],
            [0, 0, 1]
        ])
        self.translate_to_origin = np.array([
            [1, 0, center_image_meters.x],
            [0, 1, center_image_meters.y],
            [0, 0, 1]
        ])
        self.meters_to_pix = np.matmul(np.matmul(np.matmul(self.meters_to_pix, self.translate_to_origin), self.rotate_matrix), self.translate_to_image)
        self.pix_to_meters = np.linalg.inv(self.meters_to_pix)

    def __init(self):
        x_vertices = [p.x * self.args.pix_per_meter for p in self.point_cloud]
        x_delta = np.max(x_vertices) - np.min(x_vertices)
        y_vertices = [p.y * self.args.pix_per_meter for p in self.point_cloud]
        y_delta = np.max(y_vertices) - np.min(y_vertices)
        size = int(max(x_delta, y_delta))
        padding_pix = int(self.args.map_padding * size / 2)
        new_size = size + int(padding_pix * 2)
        if x_delta > y_delta:
            shift_x = padding_pix + abs(np.min(x_vertices))
            ratio = y_delta / x_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_y = padding_other + abs(np.min(y_vertices))
        else:
            shift_y = padding_pix + abs(np.min(y_vertices))
            ratio = x_delta / y_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_x = padding_other + abs(np.min(x_vertices))

        self.meters_to_pix = np.array([
            [self.args.pix_per_meter, 0, shift_x],
            [0, self.args.pix_per_meter, shift_y],
            [0, 0, 1]
        ])
        self.pix_to_meters = np.linalg.inv(self.meters_to_pix)

        self.map_image = np.zeros((new_size, new_size, 3), np.uint8)
        self.map_final = np.zeros((new_size, new_size), np.uint8)

        for point in self.point_cloud:
            vertex_pix = np.matmul(self.meters_to_pix, np.array([point.x, point.y, 1]))
            cv2.circle(self.map_image, center=(int(vertex_pix[0]), int(vertex_pix[1])), radius=0, color=(255, 255, 255), thickness=2)
        self.raw_image = copy.copy(self.map_image)
        self.__rotate_map()

    def get_map(self):
        return copy.copy(self.map_image)

    def draw_image(self, draw_path=False, draw_lines=False, lines=None):
        map_image = copy.copy(self.map_image)
        if np.any(self.path_points) and draw_path:
            for idx in range(1, len(self.path_points)):
                before_pix = np.matmul(self.meters_to_pix, np.array([self.path_points[idx-1].x, self.path_points[idx-1].y, 1]))
                after_pix = np.matmul(self.meters_to_pix, np.array([self.path_points[idx].x, self.path_points[idx].y, 1]))
                cv2.line(map_image, (int(before_pix[0]), int(before_pix[1])), (int(after_pix[0]), int(after_pix[1])), color=(0,0,255), thickness=1)

        if draw_lines:
            if lines is None:
                lines = self.lines
            for line in lines:
                cv2.line(map_image, (int(line.start_point.x), int(line.start_point.y)), (int(line.end_point.x), int(line.end_point.y)), color=(0,255,0), thickness=2)

        show_img = cv2.flip(map_image, 1)
        return show_img

    def get_point_coordinates_in_pix(self, points):
        points_out = np.zeros((len(points), 3))
        if np.any(points):
            for idx in range(1, len(points)):
                points_out[idx] = np.matmul(self.meters_to_pix, np.array([points[idx].x, points[idx].y, 1]))
        return points_out

    def rotate_and_scale_data(self):
        line_start_points_out = np.zeros((len(self.lines), 3))
        line_end_points_out = np.zeros((len(self.lines), 3))
        if np.any(self.lines):
            for idx in range(0, len(self.lines)):
                line_start_points_out[idx] = np.array([self.lines[idx].start_point.x, self.lines[idx].start_point.y, 1])
                line_end_points_out[idx] = np.array([self.lines[idx].end_point.x,self.lines[idx].end_point.y, 1])
        self.line_start_points_in_meters = line_start_points_out
        self.line_end_points_in_meters = line_end_points_out

        for idx in range(len(self.point_cloud)):
            result = np.matmul(self.meters_to_pix, np.array([self.point_cloud[idx].x,self.point_cloud[idx].y, 1]))
            self.point_cloud[idx].x = result[0]
            self.point_cloud[idx].y = result[1]
            self.point_cloud[idx].z = result[2]

        for idx in range(len(self.path_points)):
            result = np.matmul(self.meters_to_pix, np.array([self.path_points[idx].x,self.path_points[idx].y, 1]))
            self.path_points[idx].x = result[0]
            self.path_points[idx].y = result[1]
            self.path_points[idx].z = result[2]


    def plot_image(self, draw_path=False, draw_lines=False, lines=None):
        show_img = self.draw_image(draw_path, draw_lines, lines)
        cv2.imshow("Result Image", show_img)
        cv2.waitKey()
        cv2.destroyAllWindows()

    def save_image(self, path_out, draw_path=False, draw_lines=False, lines=None):
        save_img = self.draw_image(draw_path, draw_lines, lines)
        cv2.imwrite(path_out, save_img)

    def save_map(self, path_out):
        save_img = copy.deepcopy(self.map_image)
        save_img = 255 - save_img
        save_img[np.where(self.map_final)] = (0,0,255)
        save_img = cv2.flip(save_img, 1)
        cv2.imwrite(path_out, save_img)

    def save_to_h5(self, file_path):
        with h5py.File(file_path, 'w') as fp:
            if self.raw_image is not None:
                fp.create_dataset('raw_data', data=np.array(self.raw_image[:, :, 0], dtype=bool).tolist())
            if self.map_image is not None:
                fp.create_dataset('filtered_data', data=np.array(self.map_image[:, :, 0], dtype=bool).tolist())
            if self.map_final is not None:
                fp.create_dataset('final_data', data=np.array(self.map_final, dtype=bool).tolist())
            fp.create_dataset('meters_to_pix', data=self.meters_to_pix.tolist())
            fp.create_dataset('pix_to_meters', data=self.pix_to_meters.tolist())
            fp.create_dataset('lines_pixels', data=[l.as_list for l in self.lines])
            fp.create_dataset('point_cloud_meters', data=[p.as_list for p in self.point_cloud])

            fp.create_dataset('line_start_points_in_meters', data=[p.tolist() for p in self.line_start_points_in_meters])
            fp.create_dataset('line_end_points_in_meters', data=[p.tolist() for p in self.line_end_points_in_meters])

            if self.path_points:
                fp.create_dataset('path_points_meters', data=[p.as_list for p in self.path_points])
            if len(self.poi_drone) != 0:
                fp.create_dataset('poi_drone_points_meters', data=[p.as_list for p in self.poi_drone])
            if len(self.poi_target) != 0:
                fp.create_dataset('poi_target_points_meters', data=[p.as_list for p in self.poi_target])
            if len(self.poi_station) != 0:
                fp.create_dataset('poi_station_points_meters', data=[p.as_list for p in self.poi_station])

    def read_from_h5(self, file_path):
        with h5py.File(file_path, 'r') as fp:
            self.meters_to_pix = np.array(list(fp['meters_to_pix']))
            self.pix_to_meters = np.array(list(fp['pix_to_meters']))
            for line in list(fp['lines_pixels']):
                self.lines.append(Line([line[0], line[1]], [line[2], line[3]]))
            self.path_points = []

            if 'path_points_meters' in fp:
                for point in list(fp['path_points_meters']):
                    self.path_points.append(Point3D(point[0], point[1], point[2]))

            self.point_cloud = []
            for point in list(fp['point_cloud_meters']):
                self.point_cloud.append(Point3D(point[0], point[1], point[2]))

            line_start_points_in_meters = []
            for point in list(fp['line_start_points_in_meters']):
                line_start_points_in_meters.append([point[0], point[1], point[2]])
            self.line_start_points_in_meters = np.array(line_start_points_in_meters)

            line_end_points_in_meters = []
            for point in list(fp['line_end_points_in_meters']):
                line_end_points_in_meters.append([point[0], point[1], point[2]])
            self.line_end_points_in_meters = np.array(line_end_points_in_meters)

            if "poi_drone_points_meters" in fp.keys():
                for point in list(fp['poi_drone_points_meters']):
                    self.poi_drone.append(Point(point[0], point[1]))
            if "poi_target_points_meters" in fp.keys():
                for point in list(fp['poi_target_points_meters']):
                    self.poi_target.append(Point(point[0], point[1]))
            if "poi_station_points_meters" in fp.keys():
                for point in list(fp['poi_station_points_meters']):
                    self.poi_station.append(Point(point[0], point[1]))

    def process_and_add_lines(self, lines):
        map_for_coverage = np.array(np.where(self.map_image > (100,100,100), (255,255,255), (0,0,0)), dtype=np.uint8)[:,:,0]
        for idx_line, line in enumerate(lines):
            # self.plot_image(draw_lines=True, lines=[line])

            # check and fix angle
            search_angle = np.pi / 2
            if self.args.use_45deg_lines:
                search_angle = np.pi / 4
            if line.angle > 0:
                deltas = []
                for i in range(4):
                    deltas.append(line.angle - search_angle * i)
                closest_angle = search_angle * np.argmin(np.abs(deltas))
            else:
                deltas = []
                for i in range(4):
                    deltas.append(line.angle + search_angle * i)
                closest_angle = -search_angle * np.argmin(np.abs(deltas))
            delta_angle = np.min(np.abs(deltas))

            # rotate line
            if abs(delta_angle) > 0:
                line = line.rotate_line(closest_angle)

            # check and fix position
            shift_value = int(self.args.max_shift_after_rotate * self.args.pix_per_meter / 2)
            values = []
            for shift in range(-shift_value, shift_value):
                tmp_line = line.side_shift_line(shift)
                tmp_map = np.zeros_like(self.map_final)
                tmp_map = cv2.line(tmp_map, tmp_line.start_point.as_int_list, tmp_line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))
                covered_pixels = np.sum(np.logical_and(tmp_map, map_for_coverage))
                values.append(covered_pixels)
            line = line.side_shift_line(range(-shift_value, shift_value)[np.argmax(values)])

            # check for nearby lines
            is_too_close = False
            for idx_second_line in range(len(self.lines)):
                delta_threshold = self.args.min_line_distance * self.args.pix_per_meter
                start_point_delta_x = abs(line.start_point.x * np.sin(line.angle) - \
                                          self.lines[idx_second_line].start_point.x * np.sin(self.lines[idx_second_line].angle))
                start_point_delta_y = abs(line.start_point.y * np.cos(line.angle) - \
                                          self.lines[idx_second_line].start_point.y * np.cos(self.lines[idx_second_line].angle))
                end_point_delta_x = abs(line.end_point.x * np.sin(line.angle) - \
                                        self.lines[idx_second_line].end_point.x * np.sin(self.lines[idx_second_line].angle))
                end_point_delta_y = abs(line.end_point.y * np.cos(line.angle) - \
                                        self.lines[idx_second_line].end_point.y * np.cos(self.lines[idx_second_line].angle))
                if start_point_delta_x < delta_threshold and start_point_delta_y < delta_threshold and \
                        end_point_delta_x < delta_threshold and end_point_delta_y < delta_threshold:
                    # lines position
                    tmp_mask = np.zeros_like(self.map_final)
                    tmp_line = copy.deepcopy(self.lines[idx_second_line])
                    tmp_line = tmp_line.extend_line(delta_threshold, True)
                    tmp_line = tmp_line.extend_line(delta_threshold, False)
                    tmp_mask = cv2.line(tmp_mask, tmp_line.start_point.as_int_list, tmp_line.end_point.as_int_list, 255, int(2*delta_threshold))
                    if tmp_mask[line.start_point.as_int_list[1], line.start_point.as_int_list[0]] and \
                        tmp_mask[line.end_point.as_int_list[1], line.end_point.as_int_list[0]]:
                        logger.info(f'line too close to other line. SKIPPING')
                        is_too_close = True
                        break
            if is_too_close:
                continue

            # extend line
            for is_end in (False, True):
                scores = []
                step_idx = 0
                stopping_param = 0
                while True:
                    tmp_line = copy.deepcopy(line)
                    tmp_line = tmp_line.extend_line(self.args.line_extend_step * self.args.pix_per_meter * step_idx, is_end)
                    tmp_map = np.zeros_like(self.map_final)
                    tmp_map = cv2.line(tmp_map, tmp_line.start_point.as_int_list, tmp_line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))
                    covered_pixels = np.sum(np.logical_and(tmp_map, self.map_image[:,:,0]))
                    if scores:
                        if covered_pixels - scores[-1] > 10: # smarter constant maybe from percent from new added area
                            stopping_param = 0
                        else:
                            stopping_param += 1

                    if stopping_param * self.args.line_extend_step >= self.args.max_wall_gap:
                        line = line.extend_line(self.args.line_extend_step * self.args.pix_per_meter * (step_idx - stopping_param), is_end)
                        break
                    scores.append(covered_pixels)
                    step_idx += 1

            # check covered pixel density
            tmp_map = np.zeros_like(self.map_final)
            tmp_map = cv2.line(tmp_map, line.start_point.as_int_list, line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))
            covered_pixels = np.sum(np.logical_and(tmp_map, map_for_coverage))
            pixel_density = covered_pixels / line.length / int(self.args.wall_width * self.args.pix_per_meter)
            if pixel_density < self.args.min_pixel_density:
                logger.info(f'Line pixel density is {pixel_density}. SKIPPING')
                continue

            # self.plot_image(draw_lines=True, lines=[line])
            self.lines.append(line)
            # check position collision

            # draw line in final map
            self.map_final = cv2.line(self.map_final, line.start_point.as_int_list, line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))

    def split_line(self, line, inter_point, distance):
        drone_width = self.args.drone_width * self.args.pix_per_meter - distance
        tmp_line_1 = Line(inter_point, line.end_point)
        tmp_line_2 = Line(line.start_point, inter_point)
        tmp_line_1 = tmp_line_1.shorten_line(drone_width, cut_end=False)
        tmp_line_2 = tmp_line_2.shorten_line(drone_width, cut_end=True)
        new_lines = []
        for new_line in [tmp_line_1, tmp_line_2]:
            # check if line exists
            if new_line is not None:
                # check line length
                if new_line.length > self.args.min_line_length * self.args.pix_per_meter:
                    # check covered pixel density
                    tmp_map = np.zeros_like(self.map_final)
                    tmp_map = cv2.line(tmp_map, new_line.start_point.as_int_list, new_line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))
                    covered_pixels = np.sum(np.logical_and(tmp_map, np.array(np.where(self.map_image > (100,100,100), (255,255,255), (0,0,0)), dtype=np.uint8)[:,:,0]))
                    pixel_density = covered_pixels / new_line.length / int(self.args.wall_width * self.args.pix_per_meter)
                    if pixel_density > self.args.min_pixel_density:
                        new_lines.append(new_line)
        return new_lines

    def process_path(self):
        if self.is_path_points:
            path_point_list = []
            for point in self.path_points:
                point_as_list = point.as_list
                point_pixels = np.matmul(self.meters_to_pix, np.array([point_as_list[0], point_as_list[1], 1]))
                path_point_list.append(list(point_pixels[:2]))
            path_line_string = LineString(path_point_list)

            new_lines = []
            drone_width = self.args.drone_width * self.args.pix_per_meter
            for line in self.lines:
                tmp_lines = [copy.deepcopy(line)]
                # check distances
                while any([drone_width - path_line_string.distance(l.as_linestring) > 0.1 for l in tmp_lines]):
                    for idx, l in enumerate(tmp_lines):
                        distance = path_line_string.distance(l.as_linestring)
                        if drone_width - distance > 0.1:
                            p_nearest = nearest_points(path_line_string, l.as_linestring)
                            closest_val = np.inf
                            closest_idx = 0
                            for p_idx, p in enumerate(p_nearest):
                                p_distance = l.as_linestring.distance(p)
                                if p_distance < closest_val:
                                    closest_val = p_distance
                                    closest_idx = p_idx

                            inter_point = Point(p_nearest[closest_idx].x, p_nearest[closest_idx].y)
                            splitted_lines = self.split_line(l, inter_point, distance)
                            del tmp_lines[idx]
                            tmp_lines.extend(splitted_lines)
                            break

                new_lines.extend(tmp_lines)
            self.lines = new_lines
            self.map_final = np.zeros_like(self.map_final)
            for line in self.lines:
                self.map_final = cv2.line(self.map_final, line.start_point.as_int_list, line.end_point.as_int_list, 255, int(self.args.wall_width * self.args.pix_per_meter))
