from plyfile import PlyData  # pip install plyfile
import numpy as np
import sys
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from modules.helper_scripts import normalize
import cv2
import os



class PlyReader:

    @staticmethod
    def read(path):
        with open(path, 'rb') as f:
            plydata = PlyData.read(f)

        vertices_x = plydata.elements[0].data['x']
        vertices_y = plydata.elements[0].data['y']
        vertices_z = plydata.elements[0].data['z']
        vertices = np.column_stack((vertices_x, vertices_y, vertices_z))

        drone_path = path.replace('.ply', '.txt').replace('ply', 'drone_path').replace('master', 'poses')
        path_vertices = np.array([])
        if os.path.isfile(drone_path):
            with open(drone_path, 'rb') as f:
                lines = f.readlines()
                path_vertices = np.zeros((len(lines) - 2, 3))  # [x1, y1, z1]
                for idx in range(len(lines) - 2):
                    line = lines[idx].decode('utf-8')
                    line = line.split(sep=" ")
                    path_vertices[idx] = np.array([float(line[3]), float(line[7]), float(line[11])])
                pass

        return vertices, path_vertices


    @staticmethod
    def points_3D_to_image(vertices, pix_per_unit=50, path_vertices=np.array([]), show=False):
        padding = 0.1
        x_vertices = vertices[:, 0] * pix_per_unit
        x_delta = x_vertices.max() - x_vertices.min()
        y_vertices = vertices[:, 1] * pix_per_unit
        y_delta = y_vertices.max() - y_vertices.min()
        size = int(max(x_delta, y_delta))
        padding_pix = int(padding * size / 2)
        new_size = size + int(padding_pix * 2)
        if x_delta > y_delta:
            shift_x = padding_pix + abs(x_vertices.min())
            ratio = y_delta / x_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_y = padding_other + abs(y_vertices.min())
        else:
            shift_y = padding_pix + abs(y_vertices.min())
            ratio = x_delta / y_delta
            padding_other = (new_size - (size * ratio)) / 2
            shift_x = padding_other + abs(x_vertices.min())

        to_pix = np.array([
            [pix_per_unit, 0, shift_x],
            [0, pix_per_unit, shift_y],
            [0, 0, 1]
        ])
        from_pix = np.linalg.inv(to_pix)

        image = np.zeros((new_size, new_size, 3), np.uint8)

        for vertex in vertices:
            vertex_pix = np.matmul(to_pix, np.array([vertex[0], vertex[1], 1]))
            cv2.circle(image, center=(int(vertex_pix[0]), int(vertex_pix[1])),  # change only radius, min radius=1
                       radius=1, color=(255, 255, 255), thickness=-1)  # thickness=-1 ensures that the circle is filled

        output_image = image.copy()

        if path_vertices.size != 0:
            for idx in range(len(path_vertices)):
                path_vertices[idx, :] = np.matmul(to_pix, np.array([path_vertices[idx, 0], path_vertices[idx, 1], 1]))
                if idx < len(path_vertices)-1:
                    cv2.arrowedLine(image, (int(path_vertices[idx, 0]), int(path_vertices[idx, 1])),
                                    (int(path_vertices[idx + 1, 0]), int(path_vertices[idx + 1, 1])),
                                    thickness=1, color=(0, 255, 0))
        if show:
            cv2.imshow("Result Image", image)
            cv2.waitKey()
            cv2.destroyAllWindows()

        return output_image, path_vertices

    @staticmethod
    def plot_with_gt(vertices, lines):
        if 'win32' in sys.platform:
            fig, ax = plt.subplots(1,1)
            canvas = fig.canvas
        else:
            fig = plt.figure()
            canvas = FigureCanvas(fig)
        ax = fig.gca()
        ax.axis('off')
        # plt.axis('scaled')

        ax.scatter(vertices[0], vertices[1], s=1)
        for line in lines:
            xs = [line.start_point.x ,line.end_point.x]
            ys = [line.start_point.y, line.end_point.y]
            ax.plot(xs, ys)
        plt.show(block=False)

        canvas.draw()

        width, height = fig.get_size_inches() * fig.get_dpi()
        image = np.frombuffer(canvas.tostring_rgb(), dtype='uint8').reshape(int(height), int(width), 3)
        return image

    @staticmethod
    def plot_3D(vertices, show=False):
        fig = plt.figure(figsize=(10, 10))
        ax = Axes3D(fig)

        ax.scatter(vertices[:, 0], vertices[:, 1], vertices[:, 2])
        if show:
            plt.show()

    @staticmethod
    def purge_points(image):
        r = 10
        threshold_pixels = 5

        map_binary = np.array(np.where(image == (0, 0, 0), 0, 1), dtype=bool)[:, :, 0]

        new_image = map_binary.copy()
        for i in range(map_binary.shape[0]):
            for j in range(map_binary.shape[1]):
                if bool(map_binary[i, j]):
                    if np.sum(map_binary[:i-r, j-10:j+10]) > threshold_pixels and np.sum(map_binary[i+r:, j-10:j+10]) > threshold_pixels \
                            and np.sum(map_binary[i-r:i+r, :j-10]) > threshold_pixels and np.sum(map_binary[i-r:i+r, j+ 10:]) > threshold_pixels:
                        new_image[i, j] = False

        new_image = np.array(np.where(np.tile(new_image[:, :, np.newaxis], (1, 1, 3)), image, (0, 0, 0)), dtype=np.uint8)
        return new_image


if __name__ == '__main__':
    path = r'C:\Users\aigaa\Documents\GitHub\alarm_pointcloud_backend_2021_Q4\data\input\ply\master_Timon_main__P_20_D_1_1_1.ply'
    path_1 = r'C:\Users\aigaa\Documents\GitHub\alarm_pointcloud_backend_2021_Q4\data\input\ply\demo_cleaned.ply'
    vertices, _ = PlyReader.read(path)
    vertices_2, _ = PlyReader.read(path_1)
    # image = PlyReader.points_3D_to_image(vertices, True)
    # PlyReader.purge_points(image)
    # PlyReader.plot_3D(vertices, True)
    vertices_all = np.concatenate((vertices, vertices_2), axis=0)

    fig, (ax1, ax2, ax3) = plt.subplots(1, 3)

    # ax1.scatter(vertices_all[:, 0], vertices_all[:, 1])
    ax1.scatter(vertices[:, 0], vertices[:, 1], c='r')
    ax1.scatter(vertices_2[:, 0], vertices_2[:, 1], c='b')
    ax2.scatter(vertices[:, 0], vertices[:, 1])
    ax3.scatter(vertices_2[:, 0], vertices_2[:, 1])
    plt.show()
